from time import time
from unittest import mock

import responses

from sentry.identity import register
from sentry.identity.oauth2 import OAuth2Provider
from sentry.integrations.client import ApiClient, OAuth2RefreshMixin
from sentry.models import Identity, IdentityProvider
from sentry.testutils import TestCase


class ApiClientTest(TestCase):
    @responses.activate
    def test_get(self):
        responses.add(responses.GET, "http://example.com", json={})

        resp = ApiClient().get("http://example.com")
        assert resp.status_code == 200

    @responses.activate
    def test_post(self):
        responses.add(responses.POST, "http://example.com", json={})

        resp = ApiClient().post("http://example.com")
        assert resp.status_code == 200

    @responses.activate
    def test_delete(self):
        responses.add(responses.DELETE, "http://example.com", json={})

        resp = ApiClient().delete("http://example.com")
        assert resp.status_code == 200

    @responses.activate
    def test_put(self):
        responses.add(responses.PUT, "http://example.com", json={})

        resp = ApiClient().put("http://example.com")
        assert resp.status_code == 200

    @responses.activate
    def test_patch(self):
        responses.add(responses.PATCH, "http://example.com", json={})

        resp = ApiClient().patch("http://example.com")
        assert resp.status_code == 200

    @mock.patch("sentry.shared_integrations.client.base.cache")
    @responses.activate
    def test_cache_mocked(self, cache):
        cache.get.return_value = None
        responses.add(responses.GET, "http://example.com", json={"key": "value1"})
        resp = ApiClient().get_cached("http://example.com")
        assert resp == {"key": "value1"}

        key = "integration.undefined.client:a9b9f04336ce0181a08e774e01113b31"
        cache.get.assert_called_with(key)
        cache.set.assert_called_with(key, {"key": "value1"}, 900)

    @responses.activate
    def test_get_cached_basic(self):
        responses.add(responses.GET, "http://example.com", json={"key": "value1"})

        resp = ApiClient().get_cached("http://example.com")
        assert resp == {"key": "value1"}
        assert len(responses.calls) == 1

        # should still return old value
        responses.replace(responses.GET, "http://example.com", json={"key": "value2"})
        resp = ApiClient().get_cached("http://example.com")
        assert resp == {"key": "value1"}
        assert len(responses.calls) == 1

        # make sure normal get isn't impacted
        resp = ApiClient().get("http://example.com")
        assert resp == {"key": "value2"}
        assert len(responses.calls) == 2

    @responses.activate
    def test_get_cached_query_param(self):
        responses.add(responses.GET, "http://example.com?param=val", json={})
        responses.add(responses.GET, "http://example.com?param=different", json={})

        ApiClient().get_cached("http://example.com", params={"param": "val"})
        assert len(responses.calls) == 1

        ApiClient().get_cached("http://example.com", params={"param": "val"})
        assert len(responses.calls) == 1

        ApiClient().get_cached("http://example.com", params={"param": "different"})
        assert len(responses.calls) == 2

    @responses.activate
    def test_head_cached_query_param(self):
        responses.add(responses.HEAD, "http://example.com?param=val", json={})
        responses.add(responses.HEAD, "http://example.com?param=different", json={})

        ApiClient().head_cached("http://example.com", params={"param": "val"})
        assert len(responses.calls) == 1

        ApiClient().head_cached("http://example.com", params={"param": "val"})
        assert len(responses.calls) == 1

        ApiClient().head_cached("http://example.com", params={"param": "different"})
        assert len(responses.calls) == 2

    @responses.activate
    def test_default_redirect_behaviour(self):
        destination_url = "http://example.com/destination"
        destination_status = 202
        destination_headers = {"Location": destination_url}

        responses.add(responses.GET, destination_url, status=destination_status, json={})
        responses.add(responses.DELETE, destination_url, status=destination_status, json={})

        responses.add(
            responses.GET, "http://example.com/1", status=301, headers=destination_headers
        )
        resp = ApiClient().get("http://example.com/1")
        assert resp.status_code == destination_status

        # By default, non GET requests are not allowed to redirect
        responses.add(
            responses.DELETE,
            "http://example.com/2",
            status=301,
            headers=destination_headers,
            json={},
        )
        resp = ApiClient().delete("http://example.com/2")
        assert resp.status_code == 301

        responses.add(
            responses.DELETE,
            "http://example.com/3",
            status=301,
            headers=destination_headers,
            json={},
        )
        resp = ApiClient().delete("http://example.com/3", allow_redirects=True)
        assert resp.status_code == destination_status


class OAuthProvider(OAuth2Provider):
    key = "oauth"
    name = "OAuth Provider"

    def get_client_id(self):
        return "client_id"

    def get_client_secret(self):
        return "client_secret"

    def get_refresh_token_url(self):
        return "https://example.com"


class OAuth2ApiClient(ApiClient, OAuth2RefreshMixin):
    def __init__(self, identity, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.identity = identity


class OAuth2ApiClientTest(TestCase):
    def setUp(self):
        self.user = self.create_user()
        self.organization = self.create_organization()
        self.access_token = "1234567890"
        self.identity_provider_model = IdentityProvider.objects.create(type="oauth")
        register(OAuthProvider)

    @responses.activate
    def test_check_auth(self):
        new_auth = {
            "access_token": "1234567890",
            "refresh_token": "0987654321",
            "expires_in": 45678988239,
        }
        responses.add(responses.POST, "https://example.com", json=new_auth)
        identity = Identity.objects.create(
            idp=self.identity_provider_model,
            user=self.user,
            external_id="oauth_base",
            data={
                "access_token": "access_token",
                "refresh_token": "refresh_token",
                "expires": int(time()) - 3600,
            },
        )

        client = OAuth2ApiClient(identity)
        client.check_auth()

        assert client.identity.data["access_token"] == new_auth["access_token"]
        assert client.identity.data["refresh_token"] == new_auth["refresh_token"]
        assert client.identity.data["expires"] > int(time())

    @responses.activate
    def test_check_auth_no_refresh(self):
        new_auth = {
            "access_token": "1234567890",
            "refresh_token": "0987654321",
            "expires_in": 45678988239,
        }
        old_auth = {
            "access_token": "access_token",
            "refresh_token": "refresh_token",
            "expires": int(time()) + 3600,
        }
        responses.add(responses.POST, "https://example.com", json=new_auth)
        identity = Identity.objects.create(
            idp=self.identity_provider_model,
            user=self.user,
            external_id="oauth_base",
            data=old_auth,
        )

        client = OAuth2ApiClient(identity)
        client.check_auth()

        assert client.identity.data == old_auth
